import re
from measurement.measures import Weight  # type: ignore


class RaceWeight(Weight):
    """
    A thin wrapper around the measurement library Weight class to allow for the creation of Weight objects
    from strings.
    """

    REGEX = r"(?:(\d+)(?:st|\-))?(?:(\d+)(?:lb)*)?"

    def __init__(self, weight: str = None, **kwargs):
        """
        Initialize a RaceWeight object from a string.
        """

        if weight:
            st, lbs = re.match(RaceWeight.REGEX, weight).groups()
            super().__init__(self, lb=(int(st or 0) * 14 + int(lbs or 0)))  # type: ignore
        else:
            super().__init__(self, **kwargs)

    def __repr__(self) -> str:
        """
        Returns the weight as a repr.
        """
        return f"<RaceWeight: {str(self)}>"

    def __str__(self) -> str:
        """
        Returns the weight as a string.
        """
        st = int(self.lb // 14)
        lb = int(self.lb % 14)
        return f"{st}st {lb}lb"



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore


class AgeCategory(ParsingEnum):
    """
    An enumeration that represents the age category of a horse.

    """

    JUVENILE = 1
    VETERAN = 2

    # Alternatives
    VETERANS = VETERAN



from typing import Self
from .disaster import Disaster
from .finishing_position import FinishingPosition


class Outcome:
    def __init__(self, value: int | str | Disaster | FinishingPosition):
        if not isinstance(value, (Disaster, FinishingPosition)):
            if str(value).isdigit():
                value = FinishingPosition(value)
            else:
                try:
                    value = Disaster[str(value)]
                except KeyError:
                    raise ValueError(f"Invalid outcome: {value}")

        assert not isinstance(value, str)

        self._value = value

    def __repr__(self) -> str:
        return f"<Outcome: {self._value.name.title() if isinstance(self._value, Disaster) else int(self._value)}>"

    def __str__(self) -> str:
        return (
            self._value.name.title()
            if isinstance(self._value, Disaster)
            else str(self._value)
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(self._value, Disaster):
            return isinstance(getattr(other, "_value", other), Disaster)

        return self._value == getattr(other, "_value", other)

    def __lt__(self, other: Self) -> bool:
        if isinstance(self._value, Disaster):
            return not isinstance(other._value, Disaster)

        return not isinstance(other._value, Disaster) and self._value < other._value

    def __le__(self, other: Self) -> bool:
        return self == other or self < other

    def __gt__(self, other: Self) -> bool:
        if isinstance(self._value, Disaster):
            return False

        return isinstance(other._value, Disaster) or self._value > other._value

    def __ge__(self, other: Self) -> bool:
        return self == other or self > other

    @property
    def is_completion(self) -> bool:
        return isinstance(self._value, FinishingPosition)

    @property
    def is_win(self) -> bool:
        return self._value == FinishingPosition(1)



import re
from decimal import Decimal
from measurement.measures import Distance  # type: ignore


class RaceDistance(Distance):
    """
    A thin wrapper around the measurement library Distance class to allow for the creation of Distance objects
    from strings and to provide a way to initialize with furlongs.
    """

    REGEX = r"(?:(\d+)(?:m)\s*)?(?:(\d+)(?:f)\s*)?(?:(\d+)(?:y)\s*)?"

    def __init__(self, distance: str = None, **kwargs) -> None:
        """
        Initialize a RaceDistance object from a string.

        """
        if distance:
            if not re.fullmatch(r"(?:\d+[m|f|y]\s*)*", distance.replace(",", "")):
                raise AttributeError(f"Invalid distance string: {distance}")

            miles_or_metres, furlongs, yards = re.match(
                RaceDistance.REGEX, distance.replace(",", "")
            ).groups()

            if int(miles_or_metres or 0) > 10:
                kwargs["m"] = int(miles_or_metres or 0)
            else:
                kwargs["yd"] = (
                    int(miles_or_metres or 0) * 1760
                    + int(furlongs or 0) * 220
                    + int(yards or 0)
                )

        super().__init__(self, **kwargs)

    def __repr__(self) -> str:
        """
        Returns the distance as a repr.
        """
        return f"<RaceDistance: {str(self)}>"

    def __str__(self) -> str:
        """
        Returns the distance as a string.
        """
        mile = self.furlong // 8
        furlong = (self.furlong % 8) // 1
        yard = int((self.furlong % 1) * 220)
        return " ".join(
            [
                f"{mile}m" if mile else "",
                f"{furlong}f" if furlong else "",
                f"{yard}y" if yard else "",
            ]
        ).strip()

    @property
    def furlong(self) -> Decimal:
        """
        Returns the distance in furlongs.
        """
        return Decimal(self.chain / 10)



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore


class GoingDescription(ParsingEnum):
    """
    A parent enumeration for more specific going descriptions. Not intended to be instantiated directly.

    """

    pass



from peak_utility.enumeration.parsing_enum import ParsingEnum  # type: ignore


class StallsPosition(ParsingEnum):
    INSIDE = 1
    CENTRE = 2
    OUTSIDE = 3

    # Alternatives
    FAR = INSIDE
    MIDDLE = CENTRE
    NEAR = OUTSIDE



from .going_description import GoingDescription


class DirtGoingDescription(GoingDescription):
    """
    An enumeration that represents a scale of US dirt going descriptions.
    """

    MUDDY = 6
    SLOW = 7
    SLOPPY = 7.5
    GOOD = 8
    WET_FAST = 8.5
    SEALED = 9
    FAST = 10



from typing import List
from horsetalk import Disaster, FinishingPosition, FormBreak


class FormFigures:
    """
    A class to parse a string of form figures into a list of corresponding objects.

    Methods:
        parse:
            Parses the given string of form figures and returns a list of corresponding objects.

    """

    @staticmethod
    def parse(form_figures: str) -> List[FinishingPosition | FormBreak | Disaster]:
        """
        Parses the given string of form figures and returns a list of corresponding objects.
        Each form figure is converted to a corresponding object based on its value, where:
            - If the figure is a digit, it is converted to a FinishingPosition object.
            - If the figure is a valid FormBreak value, it is converted to a FormBreak object.
            - If the figure is a valid Disaster value, it is converted to a Disaster object.

        Args:
            form_figures: The string of form figures to parse.

        Returns:
            A list of objects representing the form figures.

        Raises:
            None

        """
        return [
            FinishingPosition(figure)
            if figure.isdigit()
            else FormBreak(figure)
            if figure in [member.value for member in FormBreak]
            else Disaster[figure]
            for figure in form_figures
        ]



import re
from peak_utility.number import Ordinal  # type: ignore


class FinishingPosition(Ordinal):
    """
    A class that represents the finishing position of a horse in a race.

    """

    def __new__(cls, value, *, tied=False):
        if int(value) == 0:
            if tied:
                raise ValueError("Cannot tie for unplaced")
            return int.__new__(cls, 0)

        instance = super().__new__(cls, value)
        instance.tied = tied
        return instance

    def __bool__(self):
        return int(self) >= 0

    def __repr__(self):
        return f"<FinishingPosition: {str(self)}>"

    def __str__(self):
        if int(self) == 0:
            return "Unplaced"
        return f"{'=' if self.tied else ''}{super().__repr__()}"

    @classmethod
    def parse(cls, value: int | str):
        tied = "=" in str(value)
        value = (
            re.sub(r"un*pla*c*e*d*", "", str(value).lower()).replace("=", "").strip()
        )
        return cls(value or 0, tied=tied)



from pendulum import DateTime
from .going import Going
from .race_designation import RaceDesignation
from .race_distance import RaceDistance
from .race_level import RaceLevel
from .racecourse import Racecourse
from .stalls_position import StallsPosition


class RaceConditions:
    """
    A class for grouping together race conditions into a single object.

    """

    def __init__(
        self,
        *,
        datetime: DateTime,
        racecourse: Racecourse,
        distance: RaceDistance,
        going: Going,
        race_designation: RaceDesignation,
        race_level: RaceLevel,
        stalls_position: StallsPosition | None = None,
    ):
        """
        Initialize a RaceConditions instance.

        Args:
            datetime: The datetime of the race
            racecourse: The racecourse on which the race is run
            distance: The race distance
            going: The going of the race
            race_designation: The designation of the race, i.e. whether it is a handicap, maiden, etc.
            race_level: The level of the race, i.e. Group 1, Group 2, etc.
            stalls_position: The position of the stalls on the track

        """
        self.datetime = datetime
        self.racecourse = racecourse
        self.distance = distance
        self.going = going
        self.race_designation = race_designation
        self.race_level = race_level
        self.stalls_position = stalls_position

    def __repr__(self):
        return (
            f"<RaceConditions: datetime={self.datetime}, "
            f"racecourse={self.racecourse!r}, "
            f"distance={self.distance}, "
            f"going={self.going}, "
            f"race_designation={self.race_designation.name.title()}, "
            f"race_level={self.race_level.grade or self.race_level.class_}, "
            f"stalls_position={self.stalls_position}>"
        )

    def __str__(self):
        return (
            f"{self.datetime.format('D MMM YYYY, HH:mm')}, "
            f"{self.racecourse.name}, "
            f"{self.distance.furlong}f ({self.going}), "
            f"{self.race_designation.name.title()} "
            f"{self.race_level}"
            f"{', Stalls: ' + str(self.stalls_position.name.title()) if self.stalls_position else ''}"
        )
